style = {'base_mpl_style': 'ggplot',
         'marketcolors'  : {'candle': {'up': '#000000', 'down': '#ff0000'},
                            'edge'  : {'up': '#000000', 'down': '#ff0000'},
                            'wick'  : {'up': '#606060', 'down': '#606060'},
                            'ohlc'  : {'up': '#000000', 'down': '#ff0000'},
                            'volume': {'up': '#6f6f6f', 'down': '#ff4040'},
                            'vcedge': {'up': '#1f77b4', 'down': '#1f77b4'},
         'vcdopcod'      : False,
         'alpha'         : 0.9},
         'mavcolors'     : None,
         'facecolor'     : 'w',
         'gridcolor'     : '#c0c0c0',
         'gridstyle'     : '-',
         'y_on_right'    : True,
         'rc'            : {'axes.grid.axis': 'both',
                            'axes.grid'     : True,
                            'axes.edgecolor': '#c0c0c0',
                            'axes.labelcolor': 'k',
                            'ytick.color'   : 'k',
                            'xtick.color'   : 'k',
                            'lines.markeredgecolor': 'k',
                            'patch.force_edgecolor': True},
         'base_mpf_style': 'checkers'}
